package backup

import aws.sdk.kotlin.services.s3.*
import aws.sdk.kotlin.services.s3.model.CompletedMultipartUpload
import aws.sdk.kotlin.services.s3.model.CompletedPart
import aws.sdk.kotlin.services.s3.model.GetObjectRequest
import aws.sdk.kotlin.services.s3.model.UploadPartResponse
import aws.smithy.kotlin.runtime.content.ByteStream
import aws.smithy.kotlin.runtime.content.toByteArray
import aws.smithy.kotlin.runtime.content.toInputStream
import kotlinx.coroutines.*
import ziputils.*
import ziputils.CentralDirectoryFileHeader
import ziputils.EndOfCentralDirectoryLocator
import ziputils.EndOfCentralDirectoryRecord
import ziputils.EndOfCentralDirectoryRecord64
import java.io.*
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.attribute.BasicFileAttributeView
import java.time.Instant
import java.util.zip.ZipEntry
import java.util.zip.ZipInputStream
import java.util.zip.ZipOutputStream
import kotlin.io.path.createDirectory

/**
 * AWS S3 backup client.
 */
class BackupClient(
    private val s3: S3Client,
    private val bucketName: String,
    private val bufSize: Int = 1024 * 1024 * 32,
) {
    /**
     * Upload a file/directory backup to AWS S3.
     * @param file The File object for the file or directory.
     */
    suspend fun upload(file: File) =
        coroutineScope {
            val backupKey = "${file.canonicalFile.name}/${Instant.now()}.zip"
            PipedInputStream().use { inputStream ->
                val outputStream = PipedOutputStream(inputStream)
                val zipper =
                    launch(Dispatchers.IO) {
                        file.compressToZip(outputStream)
                    }

                val data = ByteArray(bufSize)
                val initialRead = inputStream.readNBytes(data, 0, bufSize)
                if (initialRead == bufSize) {
                    // Large upload, use multipart
                    // TODO: multipart uploads can be asynchronous, which would improve
                    //       performance a little bit for big uploads.
                    val upload =
                        s3.createMultipartUpload {
                            bucket = bucketName
                            key = backupKey
                        }
                    try {
                        val uploadParts = mutableListOf<CompletedPart>()
                        var number = 1
                        var bytesRead = initialRead
                        while (bytesRead > 0) {
                            val part =
                                s3.uploadPart {
                                    bucket = bucketName
                                    key = backupKey
                                    partNumber = number
                                    uploadId = upload.uploadId
                                    body = ByteStream.fromBytes(data.take(bytesRead))
                                }.toCompletedPart(number)
                            uploadParts.add(part)
                            number++
                            bytesRead = inputStream.readNBytes(data, 0, bufSize)
                        }
                        s3.completeMultipartUpload {
                            bucket = bucketName
                            key = backupKey
                            uploadId = upload.uploadId
                            multipartUpload =
                                CompletedMultipartUpload {
                                    parts = uploadParts
                                }
                        }
                    } catch (e: Exception) {
                        s3.abortMultipartUpload {
                            bucket = bucketName
                            key = backupKey
                            uploadId = upload.uploadId
                        }
                        throw e
                    }
                } else {
                    // Small upload, use single request
                    s3.putObject {
                        bucket = bucketName
                        key = backupKey
                        body = ByteStream.fromBytes(data.take(initialRead))
                    }
                }
                zipper.join() // Should be instant
            }
            backupKey
        }

    /**
     * Restore a backup from AWS S3.
     * @param destination The destination directory path for the backup contents.
     * @param backupKey The S3 key of the backup.
     */
    suspend fun restore(
        destination: Path,
        backupKey: String,
    ) = coroutineScope {
        val req =
            GetObjectRequest {
                bucket = bucketName
                key = backupKey
            }
        s3.getObject(req) { resp ->
            ZipInputStream(
                resp.body?.toInputStream()
                    ?: throw IOException("S3 response is missing body"),
            ).use { zipStream ->
                zipStream.decompress { destination.resolve(it) }
            }
        }
    }

    /**
     * Restore a single file from a backup from AWS S3.
     * @param destination The destination directory path for the file from the backup.
     * @param backupKey The S3 key of the backup.
     * @param fileName The full name of the file to restore (including directories if it was under a subdirectory).
     */
    suspend fun restoreFile(
        destination: Path,
        backupKey: String,
        fileName: String,
    ) = coroutineScope {
        // For byte ranges refer to https://pkware.cachefly.net/webdocs/APPNOTE/APPNOTE-6.3.9.TXT
        val eocdReq =
            GetObjectRequest {
                bucket = bucketName
                key = backupKey
                // Assumption: EOCD has an empty comment
                // Assumption: Backups are at least 22 + 20 (= 42) bytes. Only COMPLETELY empty backups can be smaller,
                //             in which case this function would error anyway, so it should be fine to have this edge-case.
                range = "bytes=-${EndOfCentralDirectoryRecord.SIZE + EndOfCentralDirectoryLocator.SIZE}"
            }
        val eocdBytes =
            s3.getObject(eocdReq) { resp ->
                val bytes = resp.body?.toByteArray() ?: throw IOException("S3 response is missing body")
                bytes
            }
        val eocd = EndOfCentralDirectoryRecord.fromByteArray(eocdBytes, EndOfCentralDirectoryLocator.SIZE)
        val eocd64 =
            if (eocd.eocd64Required()) {
                val locator = EndOfCentralDirectoryLocator.fromByteArray(eocdBytes, 0)
                val eocd64Req =
                    GetObjectRequest {
                        bucket = bucketName
                        key = backupKey
                        range = "bytes=${locator.endOfCentralDirectory64Offset}-"
                    }
                s3.getObject(eocd64Req) { resp ->
                    val bytes = resp.body?.toByteArray() ?: throw IOException("S3 response is missing body")
                    EndOfCentralDirectoryRecord64.fromByteArray(bytes, 0)
                }
            } else {
                null
            }
        val cenOffset =
            if (eocd.centralDirectoryOffset == 0xffffffffU && eocd64 != null) {
                eocd64.centralDirectoryOffset
            } else {
                eocd.centralDirectoryOffset.toULong()
            }
        val censReq =
            GetObjectRequest {
                bucket = bucketName
                key = backupKey
                // We only know where to fetch until if we've also fetched EOCD64 (which isn't always the case).
                // So just over-fetch a little bit, these headers aren't that big anyway.
                range = "bytes=$cenOffset-"
            }
        val cen =
            s3.getObject(censReq) { resp ->
                val bytes = resp.body?.toByteArray() ?: throw IOException("S3 response is missing body")
                var p = 0
                while (p < bytes.size) {
                    try {
                        val cen = CentralDirectoryFileHeader.fromByteArray(bytes, p)
                        p += cen.size
                        if (cen.fileName == fileName) return@getObject cen
                    } catch (_: InvalidDataException) {
                        return@getObject null
                    }
                }
                null
            } ?: throw FileNotFoundException("File '$fileName' not found in backup")

        val localHeaderOffset =
            cen.extraFieldRecords.firstNotNullOfOrNull {
                if (it is Zip64ExtraFieldRecord && it.localHeaderOffset != null) it else null
            }?.localHeaderOffset ?: cen.localHeaderOffset.toULong()
        val compressedSize =
            cen.extraFieldRecords.firstNotNullOfOrNull {
                if (it is Zip64ExtraFieldRecord && it.compressedSize != null) it else null
            }?.compressedSize ?: cen.compressedSize.toULong()
        val req =
            GetObjectRequest {
                bucket = bucketName
                key = backupKey
                range = "bytes=$localHeaderOffset-${
                    // Over-fetch a bit so that ZipInputStream can see the next header (otherwise it EOFs, even though
                    // it knows exactly how much data is should read, so not sure why it reads ahead).
                    localHeaderOffset + cen.size.toULong() + compressedSize + CentralDirectoryFileHeader.SIZE.toULong()
                }"
            }
        s3.getObject(req) { resp ->
            ZipInputStream(
                resp.body?.toInputStream()
                    ?: throw IOException("S3 response is missing body"),
            ).use { zipStream ->
                zipStream.decompress(limit = 1) { name -> destination.resolve(name.takeLastWhile { it != '/' }) }
            }
        }
    }
}

/**
 * Convert an UploadPartResponse to a CompletedPart.
 * @param number The part number that was used for this part upload.
 * @return The CompletedPart object.
 */
internal fun UploadPartResponse.toCompletedPart(number: Int): CompletedPart {
    val part = this
    return CompletedPart {
        partNumber = number
        eTag = part.eTag
        checksumSha256 = part.checksumSha256
        checksumSha1 = part.checksumSha1
        checksumCrc32 = part.checksumCrc32
        checksumCrc32C = part.checksumCrc32C
    }
}

/**
 * Take first `n` items from the beginning of a ByteArray.
 * @param n The number of items to take.
 * @return A ByteArray of the first `n` items.
 */
internal fun ByteArray.take(n: Int) =
    if (n == size) {
        this // No copy
    } else {
        asList().subList(0, n).toByteArray() // TODO: One copy (toByteArray()), not sure how to do 0 copies here
    }

/**
 * Compress a file or directory as a ZIP file to an `OutputStream`.
 * @param outputStream The `OutputStream` to write the ZIP file contents to.
 */
internal fun File.compressToZip(outputStream: OutputStream) =
    ZipOutputStream(outputStream).use { zipStream ->
        val parentDir = this.canonicalFile.parent + "/"
        val fileQueue = ArrayDeque<File>()
        fileQueue.add(this)
        fileQueue.forEach { subFile ->
            val path = subFile.canonicalPath.removePrefix(parentDir)
            val subFiles = subFile.listFiles()
            if (subFiles != null) { // Is a directory
                val entry = ZipEntry("$path/")
                setZipAttributes(entry, subFile.toPath())
                zipStream.putNextEntry(entry)
                fileQueue.addAll(subFiles)
            } else { // Otherwise, treat it as a file
                BufferedInputStream(subFile.inputStream()).use { origin ->
                    val entry = ZipEntry(path)
                    setZipAttributes(entry, subFile.toPath())
                    zipStream.putNextEntry(entry)
                    origin.copyTo(zipStream)
                }
            }
        }
    }

/**
 * Decompress `ZipInputStream` contents to specified destination paths.
 * @param bufSize The buffer size to use for writing the decompressed files.
 * @param entryNameToPath A function to convert ZIP entry names to destination `Path`s.
 */
internal fun ZipInputStream.decompress(
    bufSize: Int = 1024 * 1024,
    limit: Int? = null,
    entryNameToPath: (String) -> Path,
) {
    var entry = this.nextEntry
    var count = 1
    while (entry != null) {
        val path = entryNameToPath(entry.name)
        if (entry.isDirectory) {
            path.createDirectory()
        } else {
            val buf = ByteArray(bufSize)
            path.toFile().outputStream().use { fileStream ->
                var bytesRead = this.read(buf)
                while (bytesRead > 0) {
                    fileStream.write(buf, 0, bytesRead)
                    bytesRead = this.read(buf)
                }
            }
        }
        applyZipAttributes(entry, path)
        // This is here, not in while loop, since we do not want to read more from the input stream.
        // But this.nextEntry will read from the input stream.
        if (limit != null && count++ >= limit) return
        entry = this.nextEntry
    }
}

/**
 * Set a `ZipEntry`'s attributes given a file's path.
 * @param entry The `ZipEntry` to set attributes of.
 * @param path The `Path` of the file to get the attributes from.
 */
internal fun setZipAttributes(
    entry: ZipEntry,
    path: Path,
) {
    try {
        val attrs = Files.getFileAttributeView(path, BasicFileAttributeView::class.java).readAttributes()
        entry.setCreationTime(attrs.creationTime())
        entry.setLastModifiedTime(attrs.lastModifiedTime())
        entry.setLastAccessTime(attrs.lastAccessTime())
    } catch (_: IOException) {
    }
}

/**
 * Set a file's attributes given a `ZipEntry`.
 * @param entry The `ZipEntry` to get the attributes from.
 * @param path The `Path` of the file to set the attributes of.
 */
internal fun applyZipAttributes(
    entry: ZipEntry,
    path: Path,
) {
    try {
        val attrs = Files.getFileAttributeView(path, BasicFileAttributeView::class.java)
        attrs.setTimes(entry.lastModifiedTime, entry.lastAccessTime, entry.creationTime)
    } catch (_: IOException) {
    }
}